
/*
 * kmtail.cpp
 *
 * Copyright (C) 2008 %{AUTHOR} <%{EMAIL}>
 */
#include "kmtail.h"
#include "kmconf.h"
#include "kmtailview.h"
#include "kmlogprops.h"

#include <QtGui/QDropEvent>
#include <QtGui/QPainter>
#include <QtGui/QPrinter>
#include <QDragEnterEvent>
#include <QDropEvent>
#include <qdebug.h>
#include <qlabel.h>
#include <qtimer.h>
#include <qclipboard.h>

#include <kconfigdialog.h>
#include <kfiledialog.h>
#include <kstatusbar.h>
#include <kapplication.h>
#include <kurl.h>
#include <kmenu.h>
#include <kmenubar.h>
#include <kfontdialog.h>
#include <kmessagebox.h>

#include <kaction.h>
#include <kactioncollection.h>
#include <kstandardaction.h>

#include <KDE/KLocale>
#include <ktoolbar.h>

#include <sys/types.h> 
#include <sys/stat.h> 
#include <unistd.h> 
#include <stdlib.h>
#include <iostream>
#include <fcntl.h>
#include <errno.h>

static  char buffer[64*1024];

using namespace std;

KmTail::KmTail( KCmdLineArgs* pArgs ): KXmlGuiWindow(),
      mpView(new KmTailView(this)),
      mpFiles(new KmFileList()),
			fTB( "filterToolBar" )

{
	setCaption( i18n( "multiple Tail" ), false );

	// accept dnd
	setAcceptDrops( true );

	// tell the KXmlGuiWindow that this is indeed the main widget
	setCentralWidget( mpView );

	// then, setup our actions
	setupActions();

	// add a status bar
	mpSBtext1 = new QLabel( i18n("Ready"), this );
	mpSBtext1->setFrameShape( QFrame::StyledPanel );
	mpSBtext1->setFrameShadow( QFrame::Sunken );
	statusBar()->addWidget(mpSBtext1);
	statusBar()->show();
	statusBar()->setToolTip( i18n("Maximum buffer size set at ") + QString::number(MAXLINES));

	// Enable save/restore of window settings, toolbars, etc.
	//setupGUI( Default, "/usr/src/dev/kde4/KmTail/src/kmtailui.rc" );
	
	setupGUI();

	// simple high-lighting system
	for ( int k = KmTail::DEBUG; k <= KmTail::EMERG; ++k )
	{
		mTags << QStringList();
		mDefTags << QStringList();
	}		
	mDefTags.replace( CRIT, QStringList( "critical" ) );
	mDefTags.replace( ERR, QStringList( "error" ) << QString( "SYSERR" ) << QString( "refused" ) );
	mDefTags.replace( WARNING, QStringList("warning" ) << QString( "kernel" ) << QString( "timeout" ) << QString( "timed out" ) );
	
	mDefPriCols << DEBUG_COL << INFO_COL << NOTICE_COL << WARNING_COL << ERR_COL
							<< CRIT_COL << ALERT_COL << EMERG_COL;
	mPriCols = mDefPriCols;
	
	mPrioTags << "<debug>" << "<info>" << "<notice>" << "<warning>" << "<err>" << "<crit>" << "<alert>" << "<emerg>";

	// setup signal/slot connections
	connect( mpView, SIGNAL(tabCloseRequested(int)), this, SLOT(closeTab(int)) );
	connect( mpView, SIGNAL(currentChanged(int)), this, SLOT(changedTab(int)) );
	connect( mpView->locTabBar(), SIGNAL(tabMoved(int, int)), this, SLOT(moveElems(int, int)) );

	mpFSW = new QFileSystemWatcher( this );
	connect( mpFSW, SIGNAL(fileChanged(const QString&)), this, SLOT(fsChange(const QString&)) );
	
	// check for file passed as argument
	if ( pArgs != NULL )
		mArgFile = pArgs->arg( 0 );
	
	// read saved options
	mUsrFont = false;
	mConfig = KGlobal::config();
	mpGenGroup = new KConfigGroup( mConfig, "General" );
	readOptions();

	if ( mpFiles->count() == 0  )                      // no file yet
	{
		fileOpen();
	}	

	if ( mpFiles->count() > 1  )
	{	
		mpView->setCurrentIndex( 0 );                    // view first tab
	}	
	else if ( mpFiles->count() == 1  )
	{	
		changedTab( 0 );																 // update status		
		mpView->tabBar()->hide();
	}
	else 		
	{
		close();                                        // not interested
	}
	
	// setup debug logging stream
/*	mpLogger = new QFile( "/home/robin/kmtail.log" );
	mpLogger->open( QIODevice::WriteOnly );
	mpDebug = new QDebug( mpLogger );*/
}

KmTail::~KmTail()
{
	//mpLogger->close();
}

bool KmTail::queryClose()
{
	saveOptions();

	while ( mpFiles->count() > 0 )
		closeTab( 0 );
	
	return true;
}

void KmTail::setupActions()
{
	QAction *openAction, *closeAction, *action;
	
	openAction = KStandardAction::open(this, SLOT(fileOpen()), actionCollection());
	openAction->setIcon( KIcon("tab-new") );
	closeAction = KStandardAction::close(this, SLOT(fileClose()), actionCollection());
	closeAction->setIcon( KIcon("tab-close") );
	KStandardAction::quit(kapp, SLOT(closeAllWindows()), actionCollection());

	KStandardAction::preferences(this, SLOT(confOptions()), actionCollection());

	mpFileReload = new KAction( KIcon("view-refresh"), i18n("&Reload"), this );
	actionCollection()->addAction( QLatin1String("reload"), mpFileReload );
	mpFileReload->setShortcut( Qt::Key_F5 );
	connect( mpFileReload, SIGNAL(triggered(bool)), this, SLOT(fileReload()) );

	mpProperties = new KAction( KIcon("document-properties"), i18n("&Properties"), this );
	actionCollection()->addAction( QLatin1String("properties"), mpProperties );
	mpProperties->setShortcut( Qt::ALT+Qt::Key_Return );
	connect( mpProperties, SIGNAL(triggered(bool)), this, SLOT(editProps()) );

	mpMenuHide = new KAction( KIcon("show-menu"), i18n("Show &Menu"), this );
	actionCollection()->addAction( QLatin1String("menuhide"), mpMenuHide );
	mpMenuHide->setShortcut( Qt::CTRL+Qt::Key_M );
	connect( mpMenuHide, SIGNAL(triggered(bool)), this, SLOT(menuHide()) );
	
	// menu option to hide menu!
	mpMenuHide->setCheckable( true );
	mpMenuHide->setChecked( menuBar()->isVisible() );

	// context menu for log entries
	mpPopup = new QMenu( mpView );
	mpPopup->addAction( mpMenuHide );
	mpPopup->addAction( openAction );
  mpPopup->addSeparator();
	mpPopup->addAction( mpProperties );
	mpPopup->addAction( mpFileReload );
  mpPopup->addSeparator();
  action = mpPopup->addAction( KIcon("edit-select-all"), QString(i18n("Select &All")) );
	action->setData( "selectall" );
	action->setShortcut( Qt::CTRL+Qt::Key_A );	
  action = mpPopup->addAction( KIcon("edit-copy"), QString(i18n("&Copy")) );
	action->setData( "copy" );
	action->setShortcut( Qt::CTRL+Qt::Key_C );	

	// extra toolbar for filter
	mpLnEdit = new KLineEdit( this );
	mpLnEdit->setClearButtonShown( true );
	mpLnEdit->setClickMessage( i18n("Filter log entries") );
	mpLnEdit->setToolTip( i18n("Enter a search argument and press Enter") );
	
	mpEditIcon = new KAction( KIcon("view-filter"), i18n("Filter:"), this );
	mpEditIcon->setToolTip( i18n("View filter toolbar") );
	actionCollection()->addAction( "filter_icon", mpEditIcon );
	
	mpEdit = new KAction( i18n("filter"), this );
 	mpEdit->setDefaultWidget( mpLnEdit );
	actionCollection()->addAction( "filter_edit", mpEdit );
	connect( mpLnEdit, SIGNAL(returnPressed(QString)), this, SLOT(editDone(QString)) );
	connect( mpLnEdit, SIGNAL(clearButtonClicked()), this, SLOT(clearEdit()) ); 

	toolBar( fTB )->show();
	
	// corner push buttons
	connect( mpView->openTB, SIGNAL(clicked()), this, SLOT(fileOpen()) ); 
	connect( mpView->closeTB, SIGNAL(clicked()), this, SLOT(fileClose()) ); 
	
	// menu for tabs
	mpTabPopup = new KMenu( mpView->tabBar() );
	mpTabPopup->addAction( KIcon("tab-detach"), i18n("&Detach tab"), this, SLOT(detachTab()) );
  mpTabPopup->addSeparator();
	mpTabPopup->addAction( KIcon("tab-close"), i18n("&Close tab"), this, SLOT(fileClose()) );
	connect( mpView->tabBar(), SIGNAL(contextMenu(int, const QPoint &)), this, SLOT(rmbTab(int, const QPoint &)) );
	mTabClickedIndex = -1;
}

void KmTail::fileOpen()
{
	// get a new log file
	QString lf = KFileDialog::getOpenFileName( KUrl("kfiledialog:///kmtail#") );
	if ( !lf.isEmpty() )
	{	
		addFile( lf );
	}	
}

void KmTail::fileClose()
{
	if ( mTabClickedIndex > -1 )
	{	
		closeTab( mTabClickedIndex );
		mTabClickedIndex = -1;
	}
	else
	{
		closeTab( mpView->currentIndex() );
	}
	
	if ( mpView->tabBar()->count() == 0 )
	{
		close();                      // no more tabs - exit
	}	
}


void KmTail::saveOptions()
{
	if ( mArgFile.isEmpty() )
		mpGenGroup->writeEntry( "files", mpFiles->fileNames() );
	
	mpGenGroup->writeEntry( "item-font", mFont.toString() );
	
	for ( int k = KmTail::DEBUG; k < KmTail::EMERG; ++k )
	{
		mpGenGroup->writeEntry( mPrioTags[k] + "-tags", mTags[k].join(",") );
		mpGenGroup->writeEntry( mPrioTags[k] + "-col", mPriCols[k] );
	}		

	mpGenGroup->writeEntry( "tstmp-col", mTstmpCol );
	mpGenGroup->sync();
}

void KmTail::readOptions()
{
	QStringList files;
	QString font;

  font = mpGenGroup->readEntry( "item-font", "" );
	if ( !font.isEmpty() )
	{	
		mFont.fromString( font );
		mUsrFont = true;
	}	
	
	for ( int k = DEBUG; k <= EMERG; ++k )
	{
		mTags.replace( k, mpGenGroup->readEntry( mPrioTags[k] + "-tags", mDefTags[k] ) ); 
		mPriCols.replace( k, mpGenGroup->readEntry( mPrioTags[k] + "-col", mDefPriCols[k] ) );
	}		

	mTstmpCol = mpGenGroup->readEntry( "tstmp-col", TSTMP_COL ); 	// light blue
	
	if ( mArgFile.isEmpty() )
		files = mpGenGroup->readEntry( "files", QStringList() );
	else
		files << mArgFile;                 // use file passed as argument
	
	if ( !files.isEmpty() )
	{	
		for ( int k = 0; k < files.count(); ++k )
		{	
			addFile( files[k] );
		}	
	}	
}

void KmTail::confOptions()
{
	KmConf opts( this );
	if ( opts.exec() == KDialog::Accepted )
	{
		saveOptions();
		
		for ( int k = 0; k < mpFiles->count(); ++k )
		{
			KmFile *pFile = mpFiles->file( k );
			fileReload( pFile );
		}	
	}	
}


void KmTail::addFile( const QString &rLogfile )
{
	if ( mpFiles->containsFile( rLogfile ) )
	{
		KMessageBox::sorry( this, rLogfile + i18n(" is already being watched.") );
		return;
	}	
	
	KmFile file;

	file.mpTW = new QTreeWidget;
	file.mpTW->setRootIsDecorated( false );
	file.mpTW->setAllColumnsShowFocus( true );
	file.mpTW->setAlternatingRowColors( true );
	file.mpTW->setContextMenuPolicy( Qt::CustomContextMenu );
	file.mpTW->setUniformRowHeights( true );
	file.mpTW->setHeaderHidden( true );
	file.mpTW->setSelectionMode( QAbstractItemView::ExtendedSelection );
	//file.mpTW->setStyleSheet("* { font-family: monospace; }");
		
  connect( file.mpTW, SIGNAL(customContextMenuRequested(const QPoint &)),
     	     this, SLOT(rmbItem(const QPoint&)) );
	
	file.mName = fileNm( rLogfile );
	file.mFileNm = rLogfile;
	file.mpPropsGroup = new KConfigGroup( mConfig, rLogfile );
	
	if ( file.mpPropsGroup->exists() )
	{
		file.mName = file.mpPropsGroup->readEntry( "name", file.mName );
		file.mPattern = file.mpPropsGroup->readEntry( "pattern", QString() );
		file.mTemplate = file.mpPropsGroup->readEntry( "template", QString() );
		file.mMsgFieldCnt = file.mpPropsGroup->readEntry( "field-cnt", 1 );
	}	
	else if ( mArgFile.isEmpty() )
	{
		KmLogProps dlg( this, &file );
		
		if ( dlg.exec() == KDialog::Rejected )
			return;
	}	
	
	file.mpTW->setColumnCount( file.mMsgFieldCnt + 1 );
	file.mNoTmplt = file.mTemplate.isEmpty();
	if ( file.mNoTmplt )
		file.mpTW->hideColumn( 0 );
	
	mpView->addTab( file.mpTW, file.mName );
	
	startFile( mpFiles->addFile( file ) );						// start tail process
	
	mpView->setCurrentWidget( file.mpTW ); 
	if ( mpFiles->count() == 1  )
		changedTab( 0 );															 	// update status		
	else
		mpView->tabBar()->show();
}

void KmTail::startFile( KmFile *pFile )
{
	pFile->mpSocket = new QFile( pFile->mFileNm.toAscii(), this );
	if ( !pFile->mpSocket->open( QIODevice::ReadOnly ) )
	{	
		KMessageBox::sorry( this, pFile->mFileNm + ": " + pFile->mpSocket->errorString() );
		return;
	}

	mpFSW->addPath( pFile->mFileNm );
	mpFiles->addFileMap(  pFile->mFileNm, pFile );
	
	skipLines( pFile );
	readFile( pFile );                               // find end of file
}

void KmTail::fsChange( const QString& path )
{
	KmFile *pFile = mpFiles->fileMap( path );
	QFileInfo info( path );
	
	if ( !info.exists() )
	{	
		pFile->mpSocket->close();
		
		pFile->mpTimer = new QTimer( this );
		pFile->mpTimer->setSingleShot( true );
		mpFiles->addObjMap( pFile->mpTimer, pFile );
		connect( pFile->mpTimer, SIGNAL(timeout()), this, SLOT(checkFileExists()) );
		pFile->mpTimer->start( 2000 );
		
		qDebug() << path << "deleted";
	}	
	else
	{
		if ( info.size() == 0 )
		{
			mpFSW->removePath( path );

			pFile->mpSocket->close();
			if ( !pFile->mpSocket->open( QIODevice::ReadOnly ) )
			{	
				KMessageBox::sorry( this, pFile->mFileNm + ": " + pFile->mpSocket->errorString() );
				return;
			}
			
			mpFSW->addPath( path );
			qDebug() << path << "recreated";
		}	
		else
		{	
			if ( !pFile->mpSocket->isOpen() )
			{	
				mpFSW->removePath( path );
				pFile->mpSocket->open( QIODevice::ReadOnly );
				mpFSW->addPath( path );
				qDebug() << path << "reopened";
			}	
			
			readFile( pFile );
		}	
	}	
}

void KmTail::checkFileExists()
{
	KmFile *pFile = mpFiles->objMap( QObject::sender() );
	mpFiles->removeObjMap( pFile->mpTimer );
	delete pFile->mpTimer;
	
	QFileInfo stat( pFile->mFileNm );
	if ( stat.exists() )
	{
		if ( !pFile->mpSocket->isOpen() )
		{	
			qDebug() << pFile->mFileNm << "simulate change";
			fsChange( pFile->mFileNm );
		}	
	}	
}

void KmTail::readFile( KmFile *pFile )
{
	QStringList lines;
	QBrush tstmp;
	int nLines;
	qint64 l;
	
	tstmp.setColor( mTstmpCol );

	while ( true )
	{	
		l = pFile->mpSocket->readLine( buffer, sizeof(buffer) );
		if ( l < 1 ) break;
		
		lines = QString( buffer ).split( '\r' );
		nLines = lines.count();
		
		for ( int k = 0; k < nLines; k++ )
		{
			QString line = lines[k];
			l = line.length();
			
			if ( k == nLines - 1 )
			{
				if ( line.mid(l - 1, 1) == QString( '\n' ) )
				{
					line = line.mid(0, l - 1);
				}
				else
				{
					mResidual += line;
					goto tidyUp;
				}
			}	
			else if ( k == 0 && !mResidual.isEmpty() )
			{
				line = mResidual + line;
				mResidual = "";
			}	
			
			QTreeWidgetItem *pItem = new QTreeWidgetItem( pFile->mpTW );
			
			if ( !pFile->mNoTmplt )
			{
				extractDateTime( pItem, pFile, line );
			}	
			
			parseFields( pItem, pFile, line );
			
			if ( mUsrFont )
				for ( int k = 0; k <= pFile->mMsgFieldCnt; ++k )
				{	
					pItem->setFont( k, mFont );
				}
				
			pItem->setForeground( 0, tstmp );

			if ( !pFile->mFilter.isEmpty() )
				if ( pItem->toolTip( pFile->mMsgFieldCnt ).indexOf( pFile->mFilter, 0, Qt::CaseInsensitive ) == -1 )
					pItem->setHidden( true );
		}	
	}				

tidyUp:

	while ( pFile->mpTW->topLevelItemCount() > MAXLINES )
		delete pFile->mpTW->takeTopLevelItem( 0 );
	
	for ( int k = pFile->mMsgFieldCnt; k > -1; --k )
	{	
		pFile->mpTW->resizeColumnToContents( k );					 // show scroll bar
	}
	
	int k = mpFiles->indexOf( *pFile );
	if ( k == mpView->currentIndex() )
		changedTab( k );																// update visible status bar
	else
	{	
		pFile->mpTW->scrollToBottom();
		mpView->setTabTextColor( k, Qt::red );					// new messages arrived
	}	
}

void KmTail::skipLines( KmFile *pFile )
{
	QFileInfo info( pFile->mFileNm );
	qint64 fsize = info.size();
	if ( fsize == 0 ) return;
	
	const int bytesToShow = MAXLINES*180;                // approx size of lines to display
	if ( bytesToShow < fsize )
	{
		if ( pFile->mpSocket->seek( fsize - bytesToShow ) )
		{
			char buf[4096];
			pFile->mpSocket->readLine( buf, sizeof(buf) );   // discard partial line
		}	
	}	
}

void KmTail::extractDateTime( QTreeWidgetItem *pItem, KmFile *pFile, QString &line )
{
	QRegExp rx( pFile->mPattern );
	int pos = rx.indexIn( line );
	
	if ( pos > -1 )
	{
		QString dt_in( rx.cap( 1 ) );
		QDateTime dateTime;
		
		if ( pFile->mTemplate == "MMM d hh:mm:ss" )
			pItem->setText( 0, dt_in );
		else
		{
			dt_in.replace( "  ", " " );                  			// kludge
			if ( pFile->mTemplate == "ssssssssss" )
				dateTime = QDateTime::fromTime_t( dt_in.toUInt() );
			else
				dateTime = QDateTime::fromString( dt_in, pFile->mTemplate );
			
			pItem->setText( 0, dateTime.toString( "MMM dd hh:mm:ss" ) );
		}
		
		line.remove( pos, rx.matchedLength() );
	}	
}


void KmTail::parseFields( QTreeWidgetItem *pItem, const KmFile *pFile, QString &line )
{
	int k;
	int tagindx = 2;
	int procindx = 1;
	int colindx = -1;
	QString ptag, ftag, proctag;
	QBrush fg; 
	
	proctag = line.section ( ' ', procindx, procindx );
	if ( proctag.endsWith("[]") )
	{
		QString oldtag = proctag;
		proctag = proctag.mid( 0, proctag.length() - 2);
		line.replace( oldtag, proctag  );
	}	

	if ( proctag.length() > 16 && pFile->mMsgFieldCnt > 2 )
		line.replace( proctag, proctag.mid(0, 16) );            // truncate excessive process-id fields
	
	ptag = line.section( ' ', tagindx, tagindx );
	colindx = mPrioTags.indexOf( ptag );
	
	if ( colindx > -1 )
	{
		line.remove( ptag + ' ' );
		ptag = ptag.mid( 1, ptag.length() - 2 );
		ftag = line.section( ' ', tagindx, tagindx );       // inspect next token
		if ( ftag[0] == '<' )
		{
			line.remove( ftag + ' ' );
			ftag = ftag.mid( 1, ftag.length() - 2 );
		}
	}
	
	for ( k = 1; k < pFile->mMsgFieldCnt; ++k )
		pItem->setText( k, line.section( ' ', k-1, k-1, QString::SectionSkipEmpty ));

	pItem->setText( k, line.section( ' ', k-1, -1, QString::SectionSkipEmpty ));

	if ( proctag.length() > 16 && pFile->mMsgFieldCnt > 2 )
		line.replace( proctag.mid(0, 16), proctag );            // reset for tooltip

	splitToolTip( line );
	pItem->setToolTip( pFile->mMsgFieldCnt, line );

	if ( colindx > -1 )
	{
		pItem->setToolTip( 0, "Facility...: " + ftag + "\nPriority..: " + ptag );
		
		fg.setColor( mPriCols[colindx] );
		for ( int k = 1; k <= pFile->mMsgFieldCnt; ++k )
			pItem->setForeground( k, fg );
	}
	else
	{			
		bool found = false;
		for ( int k = KmTail::EMERG; k >= KmTail::DEBUG; --k )
		{
			if ( mTags[k].count() > 0 )
			{	
				if ( highLightItem( pItem, mTags[k], mPriCols[k], mPrioTags[k] ) )
				{
					found = true;
					break;
				}
			}
		}		
		
		if (!found )
		{	
				pItem->setToolTip( 0, i18n("Priority..: info") );
		}
	}
}

void KmTail::splitToolTip( QString &line )
{
	for ( int i = 80; i < line.length(); i += 80 )
	{
		int p = line.indexOf( ' ', i );
		if ( p > -1 ) 
			line.replace( p, 1, '\n' );
		else 
		{
			p = line.indexOf( ',', i );
			if ( p > -1 )
				line.insert( p, '\n');
			else
				line.insert( i, '\n' );
		}	
	}	
}

void KmTail::closeTab( int k )
{
	KmFile *pFile = mpFiles->file( k );
	stopTail( pFile );	
	mpView->removeTab( k );
	delete pFile->mpTW;
	
	pFile->mpPropsGroup->writeEntry( "name", pFile->mName );
	pFile->mpPropsGroup->writeEntry( "pattern", pFile->mPattern );
	pFile->mpPropsGroup->writeEntry( "template", pFile->mTemplate );
	pFile->mpPropsGroup->writeEntry( "field-cnt", pFile->mMsgFieldCnt );
	pFile->mpPropsGroup->sync();

	mpFiles->removeFile( *pFile );

	k = mpView->currentIndex();									// where are we?
	changedTab( k );														// update status bar
	
	if ( mpView->tabBar()->count() == 1 )
	{	
		mpView->tabBar()->hide();
	}	
		
}

void KmTail::changedTab( int k )
{
	if ( (k > -1) & (k < mpFiles->count()) )
	{	
		KmFile *pFile = mpFiles->file( k );
		mpSBtext1->setText( pFile->mFileNm + ":   " + 
				QString::number( pFile->mpTW->topLevelItemCount() ) + i18n(" lines.") );
		mpLnEdit->setText( pFile->mFilter );
		mpView->setTabTextColor( k, QApplication::palette().windowText().color() );
		pFile->mpTW->scrollToBottom();
		setCaption( pFile->mFileNm );
	}
	else if ( k == -1 )
		mpSBtext1->setText( i18n("Ready") );
}

void KmTail::detachTab()
{
	if ( mpView->tabBar()->count() > 1 )
	{	
		int k = mTabClickedIndex > -1 ? mTabClickedIndex : mpView->currentIndex();
		KmFile *pFile = mpFiles->file( k );
		
		KProcess::startDetached( QString( "kmtail" ), QStringList( pFile->mFileNm ) );
		
		closeTab( k );	
		mTabClickedIndex = -1;
	}
}

void KmTail::editProps()
{
	const int k = mpView->currentIndex();
	KmFile *pFile = mpFiles->file( k );
	KmLogProps dlg( this, pFile );
	
	if ( dlg.exec() == KDialog::Accepted )
	{	
		pFile->mNoTmplt = pFile->mTemplate.isEmpty();
		if ( pFile->mNoTmplt )
			pFile->mpTW->setColumnHidden( 0, true );
		else
			pFile->mpTW->setColumnHidden( 0, false );

		mpView->setTabText( k, pFile->mName );
		fileReload( pFile );
	}	
}

void KmTail::fileReload( KmFile *pFile )
{
	if ( pFile == 0 )
	  pFile = mpFiles->file( mpView->currentIndex() );	

	stopTail( pFile );
	pFile->mpTW->clear();
	pFile->mpTW->setColumnCount( pFile->mMsgFieldCnt + 1 );
	
	startFile( pFile );
}

void KmTail::stopTail( KmFile *pFile )
{
	pFile->mpSocket->close();
	delete pFile->mpSocket;
	mpFiles->removeFileMap( pFile->mFileNm );
	mpFSW->removePath( pFile->mFileNm );
}

void KmTail::menuHide()
{
	if ( menuBar()->isVisible() )
	{
		menuBar()->hide();
		mpMenuHide->setChecked( false );
	}	
	else
	{
		menuBar()->show();
		mpMenuHide->setChecked( true );
	}	
}

void KmTail::dragEnterEvent( QDragEnterEvent *event )
{
	if ( event->mimeData()->hasUrls() )
		event->acceptProposedAction();
}

void KmTail::dropEvent( QDropEvent *event )
{
	KUrl url( event->mimeData()->text() );
	
	if ( url.scheme() == "file" )
	{	
		QString lfn( url.toLocalFile() );
		event->acceptProposedAction();
		
		addFile( lfn );
	}	
}

void KmTail::rmbItem( const QPoint &point )
{
	QAction *action;
	int k = mpView->currentIndex();
	KmFile *pFile = mpFiles->file( k );
	
	action = mpPopup->exec( pFile->mpTW->viewport()->mapToGlobal(point) );    
	
	if ( action == 0 )
		return;
	else if ( action->data().toString() == "selectall" )
		selectAllItems( pFile );
	else if ( action->data().toString() == "copy" )
		copySelection( pFile );
}

void KmTail::rmbTab( int k, const QPoint &point )
{
	mTabClickedIndex = k;                 // remember for close/detach menu opts
	mpTabPopup->exec( point );    
}

void KmTail::selectAllItems( KmFile *pFile )
{
	pFile->mpTW->selectAll();
}

void KmTail::copySelection( KmFile *pFile )
{
	QList<QTreeWidgetItem *> items = pFile->mpTW->selectedItems();
	QString lines;
	
	for ( int j = 0; j< items.count(); ++j )
	{	
		if ( pFile->mNoTmplt )
			lines += items[j]->text( pFile->mMsgFieldCnt );
		else
			lines += items[j]->text( 0 ) + ' ' + items[j]->toolTip( pFile->mMsgFieldCnt );
		lines += '\n';
	}	

	QClipboard *clipboard = QApplication::clipboard();
	clipboard->setText( lines );
}


void KmTail::editDone( QString arg )
{
	if ( arg.isEmpty() ) return;
 
	int k = mpView->currentIndex();
	KmFile *pFile = mpFiles->file( k );
	pFile->mFilter = arg;
 
	for ( int j = 0; j < pFile->mpTW->topLevelItemCount(); ++j )
	{
		if ( pFile->mpTW->topLevelItem( j )->toolTip( pFile->mMsgFieldCnt ).indexOf( arg, 0, Qt::CaseInsensitive ) == -1 )
			pFile->mpTW->topLevelItem( j )->setHidden( true );
	}
}

void KmTail::clearEdit()
{
	int k = mpView->currentIndex();
	KmFile *pFile = mpFiles->file( k );
	pFile->mFilter.clear();
 
	for ( int j = 0; j < pFile->mpTW->topLevelItemCount(); ++j )
	{
		if ( pFile->mpTW->topLevelItem( j )->isHidden() )
			pFile->mpTW->topLevelItem( j )->setHidden( false );
	}
	
	pFile->mpTW->scrollToBottom();
}

bool KmTail::highLightItem( QTreeWidgetItem *pItem , QStringList &tags , QString &col, const QString &prio ) const
{
	QBrush brush;
	bool found = false;
	const int ncols = pItem->columnCount();
	
	brush.setColor( col );
	
	for ( int j = 0; j < tags.count(); ++j )
	{
		if ( pItem->toolTip( ncols - 1 ).indexOf( tags[j], 0, Qt::CaseInsensitive ) >= 0 )
		{	
			for ( int k = 1; k < ncols; ++k )
			{	
				pItem->setForeground( k, brush );
			}
				
			pItem->setToolTip( 0, i18n("Priority..: ") + prio.mid( 1, prio.length() - 2 ) );
			found = true;
			break;
		}
	}	
	
	return found;
}

void KmTail::moveElems( int from, int to )
{
	mpFiles->move( from, to );
}

QString KmTail::fileNm( const QString &rLogfile )
{
	QString name;
	int p = rLogfile.lastIndexOf( '/' );
	name = rLogfile.mid( p+1 );
	
	if ( name.right( 4 ) == ".log" )
		name.chop( 4 );
	else if ( name.right( 5 ) == ".fifo" )
		name.chop( 5 );
	else if ( name.right( 5 ) == ".pipe" )
		name.chop( 5 );

	return name;
}

#include "kmtail.moc"
